//
// $History: Blowfish.pas $
// 
// *****************  Version 6  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
//
// *****************  Version 3  *****************
// User: Lvk          Date: 22.07.98   Time: 17:19
// Updated in $/Delphi 4.0/StreamFilter
// Fixed some comment blocks
//
// *****************  Version 2  *****************
// User: Lvk          Date: 21.07.98   Time: 23:50
// Updated in $/Delphi 4.0/StreamFilter
// Made the base class (TCryptography) expand the key and make a binary
// key of the correct size, thereby reducing code in each descendant
// class.
//
// *****************  Version 1  *****************
// User: Lvk          Date: 19.07.98   Time: 23:47
// Created in $/Delphi 4.0/StreamFilter
// Blowfish encryption algorithm, developed by Bruce Schneier
//

unit Blowfish;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net
//
// Blowfish is a public domain, unpatented, encryption algorithm, developed by
// Bruce Schneier, 1994.

interface

uses
	Cryptography, BlowfishData, AuxFuncs;

type
	TBlowfishBlock	= array[1..2] of LongWord;

{ TBlowfish class }
	TBlowfish	= class(TCryptography)
	private
		P	: TPType;
		S	: TSType;
		procedure Cipher(var Block: TBlowfishBlock; StartN, DeltaN: Integer); virtual;
	protected
		procedure CalculateSubKeys; override;
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create; overload;
		constructor Create(AKey: string); overload;
{$ELSE}
// Delphi 3.x specific code
		constructor Create;
		constructor CreateKey(AKey: string);
{$ENDIF}
		destructor Destroy; override;

		function EncipherBlock(var Block): Boolean; override;
		function DecipherBlock(var Block): Boolean; override;
	end; // class TBlowfish

implementation

uses
	SysUtils;
	
{ TBlowfish }

procedure TBlowfish.CalculateSubKeys;
var
	i,j			: Integer;
	Temp		: LongWord;
	Block		: TBlowfishBlock;
begin
	// Copy initial data to local arrays
	P := InitP;
	S := InitS;

	// Xor the key into the P array
	for i := 0 to Rounds + 1 do
	begin
		Move((KeyPtr+i*4)^, Temp, 4);
		P[i] := P[i] xor Temp;
	end; // for i := 0 to Rounds + 1

	// Now encrypt a 0-data block N times to make new P and S tables
	Block[1] := 0;
	Block[2] := 0;

	i := 0;
	while (i < Rounds + 1) do
	begin
		Cipher(Block, 0, 1);
		P[i] := Block[1];
		P[i+1] := Block[2];
		Inc(i, 2);
	end; // while (i < Rounds + 1)

	for j := 0 to 3 do
	begin
		i := 0;
		while (i < 256) do
		begin
			Cipher(Block, 0, 1);
			S[j, i] := Block[1];
			S[j, i+1] := Block[2];
			Inc(i, 2);
		end; // while (i < Rounds + 1)
	end; // for j := 0 to 3
end; // procedure TBlowfish.CalculateSubKeys

{$IFDEF VER120}
// Delphi 4.x specific code
constructor TBlowfish.Create(AKey: string);
begin
	inherited Create('Blowfish', 8, (Rounds + 2) * 4, AKey);
end; // constructor TBlowfish.Create(String)
{$ELSE}
// Delphi 3.x specific code
constructor TBlowfish.CreateKey(AKey: string);
begin
	inherited CreateKey('Blowfish', 8, (Rounds + 2) * 4, AKey);
end; // constructor TBlowfish.Create(String)
{$ENDIF}

constructor TBlowfish.Create;
begin
	inherited Create('Blowfish', 8, (Rounds + 2) * 4);
end; // constructor TBlowfish.Create

function TBlowfish.DecipherBlock(var Block): Boolean;
begin
	// Call the cipher function
	Cipher(TBlowfishBlock(Block), Rounds + 1, -1);

	// This function cannot fail, unless there is an exception
	Result := TRUE;
end; // function TBlowfish.DecipherBlock

destructor TBlowfish.Destroy;
begin
	inherited Destroy;
end; // destructor TBlowfish.Destroy

function TBlowfish.EncipherBlock(var Block): Boolean;
begin
	// Call the cipher function
	Cipher(TBlowfishBlock(Block), 0, 1);

	// This function cannot fail, unless there is an exception
	Result := TRUE;
end; // function TBlowfish.EncipherBlock

{$R-,Q-}
procedure TBlowfish.Cipher(var Block: TBlowfishBlock; StartN,
	DeltaN: Integer);
var
	I,N		: Integer;
	Xl,Xr	: LongWord;
	Temp	: LongWord;
	Bytes	: record
		d,c,b,a	: Byte;
	end absolute Xl;
	F			: LongWord;
begin
	// Get the data to encipher
	Xl := Block[1];
	Xr := Block[2];

	N := StartN;
	for i := 0 to Rounds - 1 do
	begin
		// Do the math
		Xl := Xl xor P[N];
		Inc(N, DeltaN);
{$IFDEF VER120}
// Delphi 4.x specific code
		F := ((S[0, Bytes.a]+(S[1, Bytes.b] mod 32)) xor S[2, Bytes.c]) + (S[3, Bytes.d] mod 32);
{$ELSE}
// Delphi 3.x specific code
		F := ((S[0, Bytes.a]+((S[1, Bytes.b] mod 32+32) mod 32)) xor S[2, Bytes.c]) + ((S[3, Bytes.d] mod 32+32) mod 32);
{$ENDIF}
		Xr := Xr xor F;

		// Now prepare for next round
		Temp := Xl;
		Xl := Xr;
		Xr := Temp;
	end; // for i := 0 to Rounds - 1

	// Swap an additional time
	Temp := Xl;
	Xl := Xr;
	Xr := Temp;

	// Now do the final math
	Xr := Xr xor P[N];
	Xl := Xl xor P[N + DeltaN];

	// And return the result
	Block[1] := Xl;
	Block[2] := Xr;
end; // procedure TBlowfish.Cipher
{$R+,Q+}

end. // unit Blowfish
