//
// $History: Cryptography.pas $
// 
// *****************  Version 9  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 8  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 7  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
// 
// *****************  Version 6  *****************
// User: Lvk          Date: 22.07.98   Time: 17:20
// Updated in $/Delphi 4.0/StreamFilter
// Fixed the resourcestring
// Fixed some comment blocks
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 21.07.98   Time: 23:51
// Updated in $/Delphi 4.0/StreamFilter
// Made the base class (TCryptography) expand the key and make a binary
// key of the correct size, thereby reducing code in each descendant
// class.
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 20.07.98   Time: 20:50
// Updated in $/Delphi 4.0/StreamFilter
// Added documentation for first release
// 
// *****************  Version 3  *****************
// User: Lvk          Date: 17.07.98   Time: 23:51
// Updated in $/Delphi 4.0/StreamFilter
// Added copyright message
// 
// *****************  Version 2  *****************
// User: Lvk          Date: 17.07.98   Time: 20:44
// Updated in $/Delphi 4.0/Cryptography
// Nothing done
// 
// *****************  Version 1  *****************
// User: Lvk          Date: 17.07.98   Time: 20:43
// Created in $/Delphi 4.0/Cryptography
// Cryptography base class unit
//

unit Cryptography;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net

interface

uses
	AuxFuncs, Math, Classes;

type
	TIOFunc				= function(var Data; Size: Integer): Integer of object;
	TProcessFunc	= function(var Block): Boolean of object;

{ TCryptography class }
	TCryptography = class
	private
		FBlockSize		: Integer;
		FKey					: string;
		FBufferSize		: Integer;
		FKeySize			: Integer;
		FKeyPtr				: PChar;
		FName					: string;

		{ For use with EncipherString/DecipherString }
		String1				: string;
		StringPos			: Integer;
		String2				: string;

		{ For use with EncipherStream/DecipherStream }
		Stream1				: TStream;
		Stream2				: TStream;

		procedure SetKey(const Value: string);

		function GetString(var Data; Size: Integer): Integer; virtual;
		function PutString(var Data; Size: Integer): Integer; virtual;

		function GetStream(var Data; Size: Integer): Integer; virtual;
		function PutStream(var Data; Size: Integer): Integer; virtual;

		function ProcessData(GetData, PutData: TIOFunc; ProcessBlock: TProcessFunc): Integer; virtual;
		procedure SetBufferSize(const Value: Integer); virtual;
		procedure Init(AName: string; ABlockSize,AKeySize: Integer; AKey: string); virtual;
		procedure InvolveKey; virtual;
	protected
		procedure CalculateSubKeys; virtual; abstract;
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create(AName: string; ABlockSize,AKeySize: Integer); overload;
		constructor Create(AName: string; ABlockSize,AKeySize: Integer; AKey: string); overload;
{$ELSE}
// Delphi 3.x specific code
		constructor Create(AName: string; ABlockSize,AKeySize: Integer);
		constructor CreateKey(AName: string; ABlockSize,AKeySize: Integer; AKey: string);
{$ENDIF}
		destructor Destroy; override;

		function EncipherBlock(var Block): Boolean; virtual; abstract;
		function DecipherBlock(var Block): Boolean; virtual; abstract;

		function EncipherString(var S: string): Integer; virtual;
		function DecipherString(var S: string): Integer; virtual;

		function EncipherStream(SourceStream,DestinationStream: TStream): Integer; virtual;
		function DecipherStream(SourceStream,DestinationStream: TStream): Integer; virtual;

		function EncipherData(GetData,PutData: TIOFunc): Integer; virtual;
		function DecipherData(GetData,PutData: TIOFunc): Integer; virtual;

		property BlockSize: Integer read FBlockSize;
		property Key: string read FKey write SetKey;
		property BufferSize: Integer read FBufferSize write SetBufferSize;

		property KeyPtr: PChar read FKeyPtr;
		property KeySize: Integer read FKeySize;

		property Name: string read FName;
	end; // class TCryptography

implementation

resourcestring
	SSecret	= 'secret';
	
{ TCryptography }

constructor TCryptography.Create(AName: string; ABlockSize,AKeySize: Integer);
begin
	Init(AName, ABlockSize, AKeySize, SSecret);
end; // constructor TCryptography.Create

{$IFDEF VER120}
// Delphi 4.x specific code
constructor TCryptography.Create(AName: string; ABlockSize,AKeySize: Integer; AKey: string);
{$ELSE}
// Delphi 3.x specific code
constructor TCryptography.CreateKey(AName: string; ABlockSize,AKeySize: Integer; AKey: string);
{$ENDIF}
begin
	Init(AName, ABlockSize, AKeySize, AKey);
end; // constructor TCryptography.Create

procedure TCryptography.Init(AName: string; ABlockSize,AKeySize: Integer; AKey: string);
begin
	FName				:= AName;
	FBlockSize 	:= ABlockSize;
	FBufferSize := ABlockSize * (2048 div ABlockSize);
	FKey			 	:= SSecret;
	FKeySize		:= AKeySize;
	GetMem(FKeyPtr, AKeySize);
	FillChar(FKeyPtr^, AKeySize, #0);
	Key := AKey;
end; // procedure TCryptography.Init

function TCryptography.DecipherData(GetData, PutData: TIOFunc): Integer;
begin
	Result := ProcessData(GetData, PutData, DecipherBlock);
end; // function TCryptography.DecipherData

function TCryptography.DecipherStream(SourceStream,
	DestinationStream: TStream): Integer;
begin
	Stream1 := SourceStream;
	Stream2 := DestinationStream;

	Result := DecipherData(GetStream, PutStream);
end; // function TCryptography.DecipherStream

function TCryptography.DecipherString(var S: string): Integer;
begin
	String1 := S;
	String2 := '';
	StringPos := 1;

	Result := DecipherData(GetString, PutString);

	S := String2;
end; // function TCryptography.DecipherString

destructor TCryptography.Destroy;
begin
	FreeMem(FKeyPtr, FKeySize);
end; // destructor TCryptography.Destroy

function TCryptography.EncipherData(GetData, PutData: TIOFunc): Integer;
begin
	Result := ProcessData(GetData, PutData, EncipherBlock);
end; // function TCryptography.EncipherData

function TCryptography.EncipherStream(SourceStream,
	DestinationStream: TStream): Integer;
begin
	Stream1 := SourceStream;
	Stream2 := DestinationStream;

	Result := EncipherData(GetStream, PutStream);
end; // function TCryptography.EncipherStream

function TCryptography.EncipherString(var S: string): Integer;
begin
	String1 := S;
	String2 := '';
	StringPos := 1;

	Result := EncipherData(GetString, PutString);

	S := String2;
end; // function TCryptography.EncipherString

function TCryptography.GetStream(var Data; Size: Integer): Integer;
begin
	Result := Stream1.Read(Data, Size);
end; // function TCryptography.GetStream

function TCryptography.GetString(var Data; Size: Integer): Integer;
var
	Left		: Integer;
	ToCopy	: Integer;
begin
	if (StringPos > Length(String1)) then
		Result := 0
	else begin
		Left := (Length(String1) - StringPos) + 1;
		ToCopy := Min(Size, Left);
		Move(String1[StringPos], Data, ToCopy);
		Result := ToCopy;
		Inc(StringPos, ToCopy);
	end; // if (StringPos > Length(String1))
end; // function TCryptography.GetString

function TCryptography.ProcessData(GetData, PutData: TIOFunc;
	ProcessBlock: TProcessFunc): Integer;
var
	Buffer				: PChar;
	InBuffer			: Integer;
	BytesWritten	: Integer;
	OK						: Boolean;
	More					: Boolean;
	Error					: Boolean;
	ModifiedSize	: Integer;

	function ProcessBlocks(Buffer: PChar; Size: Integer; var ModifiedSize: Integer): Boolean;
	var
		BlockCount	: Integer;
		BlockPtr		: PChar;
		Block				: Integer;
	begin
		Result := TRUE;
		BlockCount := (Size div BlockSize) + Ord((Size mod BlockSize) > 0);
		ModifiedSize := (BlockSize * BlockCount);

		Block := 0;
		BlockPtr := Buffer;
		while (Block < BlockCount) and Result do
		begin
			Result := ProcessBlock(BlockPtr^);
			Inc(BlockPtr, BlockSize);
			Inc(Block);
		end; // while (Block < BlockCount)
	end; // function TCryptography.ProcessData.ProcessBlocks

begin
	Result := 0;
	Buffer := nil;

	try
		GetMem(Buffer, BufferSize);

		repeat
  		FillChar(Buffer^, BufferSize, 0);
			InBuffer := GetData(Buffer^, BufferSize);

			OK := ProcessBlocks(Buffer, InBuffer, ModifiedSize);

			BytesWritten := PutData(Buffer^, ModifiedSize);

			Error := (not OK) or (ModifiedSize <> BytesWritten);
			More := (InBuffer > 0);
		until (not More) and (not Error); // repeat
	finally
		FreeMem(Buffer, BufferSize);
	end; // try
end; // function TCryptography.ProcessData

function TCryptography.PutStream(var Data; Size: Integer): Integer;
begin
	Result := Stream2.Write(Data, Size);
end; // function TCryptography.PutStream

function TCryptography.PutString(var Data; Size: Integer): Integer;
var
	P				: PChar;
	S				: ShortString;
	ToCopy	: Integer;
begin
	Result := 0;

	P := @Data;
	while (Size > 0) do
	begin
		ToCopy := Min(Size, 255);
		Move(P^, S[1], ToCopy);
		S[0] := Chr(ToCopy);
		String2 := String2 + S;

		Inc(P, ToCopy);
		Dec(Size, ToCopy);
		Inc(Result, ToCopy);
	end; // while (Size > 0)
end; // function TCryptography.PutString

procedure TCryptography.SetBufferSize(const Value: Integer);
var
	TempValue	: Integer;
begin
	// Calculate size that matches block size
	TempValue := (Value div BlockSize) * BlockSize;

	// Below minimum block size?
	if (TempValue < 256) then
		TempValue := (256 div BlockSize) * BlockSize;

	// Now check if we need to alter it
	if (TempValue <> FBufferSize) then
	begin
		FBufferSize := TempValue;
	end; // if (Value <> FBufferSize)
end; // procedure TCryptography.SetBufferSize

procedure TCryptography.SetKey(const Value: string);
begin
	if (FKey <> Value) and (Value <> '') then
	begin
		FKey := Value;
		InvolveKey;
		CalculateSubKeys;
	end; // if (FKey <> Value)
	FKey := Value;
end; // procedure TCryptography.SetKey

procedure TCryptography.InvolveKey;
var
	TempKey	: string;
	i,j			: Integer;
	K1,K2		: LongWord;
begin
	// First expand the key to a multiple of KeySize bytes
	TempKey := Key;
	i := 1;
	while ((Length(TempKey) mod KeySize) <> 0) do
	begin
		TempKey := TempKey + TempKey[i];
		Inc(i);
	end; // while ((Length(TempKey) mod KeySize) <> 0)

	// Now shorten the key down to one KeySize block by combining the bytes
	i := 1;
	j := 0;
	while (i < Length(TempKey)) do
	begin
		Move((FKeyPtr+j)^, K1, 4);
		Move(TempKey[i], K2, 4);
		K1 := ROL(K1, K2) xor K2;
		Move(K1, (FKeyPtr+j)^, 4);
		j := (j + 4) mod KeySize;
		Inc(i, 4);
	end; // while (i < Length(TempKey))
end; // procedure TCryptography.InvolveKey

end. // unit Cryptography
