//
// $History: CryptographyStream.pas $
// 
// *****************  Version 7  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 6  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 21.07.98   Time: 23:51
// Updated in $/Delphi 4.0/StreamFilter
// Removed comment block documenting use
// 
// *****************  Version 3  *****************
// User: Lvk          Date: 20.07.98   Time: 20:50
// Updated in $/Delphi 4.0/StreamFilter
// Added documentation for first release
// 
// *****************  Version 2  *****************
// User: Lvk          Date: 19.07.98   Time: 23:54
// Updated in $/Delphi 4.0/StreamFilter
// Fixed author messages
// 
// *****************  Version 1  *****************
// User: Lvk          Date: 17.07.98   Time: 23:52
// Created in $/Delphi 4.0/StreamFilter
// Basic cryptography stream filter class.
// 
// Must be overriden by descendant classes implementing specific
// encryption algorithms.
//

unit CryptographyStream;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net

interface

uses
	Math, SysUtils, Classes, StreamFilters, Cryptography, AuxFuncs;

type
	TBasicCryptographyStream	= class(TBasicStreamFilter)
	private
		FEncryptionClass			: TCryptography;
		FFreeEncryptionClass	: Boolean;
		OneBlock							: PChar;
		InBlock								: Integer;
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create(ACoreStream: TStream; AEncryptionClass: TCryptography;
			DoFreeCore: Boolean=FALSE; DoFreeEncryptionClass: Boolean=TRUE);
{$ELSE}
// Delphi 3.x specific code
		constructor Create(ACoreStream: TStream; AEncryptionClass: TCryptography;
			DoFreeCore, DoFreeEncryptionClass: Boolean);
{$ENDIF}
		destructor Destroy; override;

		property EncryptionClass: TCryptography read FEncryptionClass;
		property FreeEncryptionClass: Boolean read FFreeEncryptionClass;
	end; // class TBasicCryptographyStream

	TEncryptionStream = class(TBasicCryptographyStream)
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create(ACoreStream: TStream; AEncryptionClass: TCryptography;
			DoFreeCore: Boolean=FALSE; DoFreeEncryptionClass: Boolean=TRUE);
{$ELSE}
// Delphi 3.x specific code
		constructor Create(ACoreStream: TStream; AEncryptionClass: TCryptography;
			DoFreeCore, DoFreeEncryptionClass: Boolean);
{$ENDIF}
		destructor Destroy; override;

		function Write(const Buffer; Count: Longint): Longint; override;
	end; // class TEncryptionStream

	TDecryptionStream = class(TBasicCryptographyStream)
	private
		BlockHead	: PChar;
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create(ACoreStream: TStream; AEncryptionClass: TCryptography;
			DoFreeCore: Boolean=FALSE; DoFreeEncryptionClass: Boolean=TRUE);
{$ELSE}
// Delphi 3.x specific code
		constructor Create(ACoreStream: TStream; AEncryptionClass: TCryptography;
			DoFreeCore, DoFreeEncryptionClass: Boolean);
{$ENDIF}
		destructor Destroy; override;

		function Seek(Offset: Longint; Origin: Word): Longint; override;
		function Read(var Buffer; Count: Longint): Longint; override;
	end; // class TEncryptionStream

implementation

resourcestring
	SInvalidOrigin = 'Invalid seek origin';

{ TBasicCryptographyStream }

constructor TBasicCryptographyStream.Create(ACoreStream: TStream;
	AEncryptionClass: TCryptography; DoFreeCore,
	DoFreeEncryptionClass: Boolean);
begin
	inherited Create(ACoreStream, DoFreeCore);
	FEncryptionClass := AEncryptionClass;
	GetMem(OneBlock, EncryptionClass.BlockSize);
	FFreeEncryptionClass := DoFreeEncryptionClass;
	InBlock := 0;
end; // constructor TBasicCryptographyStream.Create

destructor TBasicCryptographyStream.Destroy;
begin
	FreeMem(OneBlock, EncryptionClass.BlockSize);
	if FreeEncryptionClass then
		EncryptionClass.Free;
	inherited;
end; // destructor TBasicCryptographyStream.Destroy

{ TEncryptionStream }

constructor TEncryptionStream.Create(ACoreStream: TStream;
  AEncryptionClass: TCryptography; DoFreeCore,
	DoFreeEncryptionClass: Boolean);
begin
	inherited;
end; // constructor TEncryptionStream.Create

destructor TEncryptionStream.Destroy;

	procedure FillBlock;
	var
		Dummy	: Byte;
	begin
		Dummy := 0;
		while (InBlock > 0) do
			if (Write(Dummy, 1) <> 1) then
				Break;
	end; // procedure TEncryptionStream.Destroy.FillBlock

begin
	if InBlock > 0 then
		FillBlock;
	inherited;
end; // destructor TEncryptionStream.Destroy

function TEncryptionStream.Write(const Buffer; Count: Integer): Longint;
var
	P				: PChar;
	ToCopy	: Integer;
	Written	: Integer;
begin
	P := @Buffer;
	Result := 0;
	while (Count > 0) do
	begin
		ToCopy := Min(EncryptionClass.BlockSize - InBlock, Count);
		Move(P^, (OneBlock + InBlock)^, ToCopy);
		Inc(P, ToCopy);
		Dec(Count, ToCopy);
		Inc(Result, ToCopy);
		Inc(InBlock, ToCopy);

		if (InBlock = EncryptionClass.BlockSize) then
		begin
			if (not EncryptionClass.EncipherBlock(OneBlock^)) then
				Break;
			Written := CoreStream.Write(OneBlock^, EncryptionClass.BlockSize);
			if (Written <> EncryptionClass.BlockSize) then
				Break;

			InBlock := 0;
		end; // if (InBlock = EncryptionClass.BlockSize)
	end; // while (Count > 0)
end; // function TEncryptionStream.Write

{ TDecryptionStream }

constructor TDecryptionStream.Create(ACoreStream: TStream;
	AEncryptionClass: TCryptography; DoFreeCore,
	DoFreeEncryptionClass: Boolean);
begin
	inherited;
	InBlock := 0;
	BlockHead := OneBlock;
end; // constructor TDecryptionStream.Create

destructor TDecryptionStream.Destroy;
begin
	inherited;
end; // destructor TDecryptionStream.Destroy

function TDecryptionStream.Read(var Buffer; Count: Integer): Longint;
var
	P				: PChar;
	ToCopy	: Integer;
begin
	P := @Buffer;
	Result := 0;
	while (Count > 0) do
	begin
		if (InBlock = 0) then
		begin
			InBlock := CoreStream.Read(OneBlock^, EncryptionClass.BlockSize);
			if (InBlock <> EncryptionClass.BlockSize) then
				Break;

			if (not EncryptionClass.DecipherBlock(OneBlock^)) then
				Break;
			BlockHead := OneBlock;
			InBlock := EncryptionClass.BlockSize;
		end; // if (InBlock = 0)

		ToCopy := Min(Count, InBlock);
		Move(BlockHead^, P^, ToCopy);
		Inc(P, ToCopy);
		Inc(Result, ToCopy);
		Dec(Count, ToCopy);
		Inc(BlockHead, ToCopy);
		Dec(InBlock, ToCopy);
	end; // while (Count > 0)
end; // function TDecryptionStream.Read

function TDecryptionStream.Seek(Offset: Integer; Origin: Word): Longint;

	function SeekFromBeginning(Offset: Integer): Integer;
	var
		Dummy		: Byte;
		I				: Integer;
		ToRead	: Integer;
	begin
		// Calculate how much to read after repositioning
		ToRead := (Offset mod EncryptionClass.BlockSize);
		Dec(Offset, ToRead);

		// Do the actual seek and reset input buffers
		Result := CoreStream.Seek(Offset, soFromBeginning);
		InBlock := 0;
		BlockHead := OneBlock;

		// Adjust for real position
		for i := 1 to ToRead do
			if (Read(Dummy, 1) <> 1) then
				Break
			else
				Inc(Result);
	end; // function TDecryptionStream.Seek.SeekFromBeginning

	function SeekFromCurrent(Offset: Integer): Integer;
	var
		Dummy		: Byte;
		I				: Integer;
		ToRead	: Integer;
		RealPos	: Integer;
	begin
		if (Offset = 0) then
			Result := CoreStream.Seek(0, soFromCurrent)
		else begin
			// Calculate adjusted position
			RealPos := (Position + Offset);
			ToRead := (RealPos mod EncryptionClass.BlockSize);
			Dec(RealPos, ToRead);
			Offset := (RealPos - Position);

			// Do the actual seek and reset input buffers
			Result := CoreStream.Seek(Offset, soFromBeginning);
			InBlock := 0;
			BlockHead := OneBlock;

			// Adjust for real position
			for i := 1 to ToRead do
				if (Read(Dummy, 1) <> 1) then
					Break
				else
					Inc(Result);
		end; // if (Offset = 0) ... else
	end; // function TDecryptionStream.Seek.SeekFromCurrent

	function SeekFromEnd(Offset: Integer): Integer;
	var
		StreamSize	: Integer;
	begin
		StreamSize := CoreStream.Seek(0, soFromEnd);
		Result := SeekFromBeginning(StreamSize + Offset);
	end; // function TDecryptionStream.Seek.SeekFromEnd

begin
	case Origin of
		soFromBeginning	: Result := SeekFromBeginning(Offset);
		soFromCurrent		: Result := SeekFromCurrent(Offset);
		soFromEnd				: Result := SeekFromEnd(Offset);
	else
		raise Exception.Create(SInvalidOrigin);
	end; // case Origin
end; // function TDecryptionStream.Seek

end. // unit CryptographyStream

