//
// $History: Logging.pas $
// 
// *****************  Version 6  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 12.09.98   Time: 17:40
// Updated in $/Delphi 4.0/Indexer
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 3  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
// 
// *****************  Version 2  *****************
// User: Administrator Date: 29.08.98   Time: 13:31
// Updated in $/Delphi 4.0/Indexer
// Changed email address
// 
// *****************  Version 1  *****************
// User: Lvk          Date: 24.07.98   Time: 1:17
// Created in $/Delphi 4.0/StreamFilter
// Logging class, for debugging purposes
//

unit Logging;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net

interface

var
	LogFilename				: string	= '';
	IncludeTimeStamp	: Boolean	= FALSE;
	AutoFlush					: Boolean	= FALSE;
	LineLength				: Integer	= 80;

{$DEFINE LOGGING}
{$DEFINE BLANKLOG}
// If BLANKLOG is defined, then the log file will be erased when the program
// starts.

// A blank message will make a totally blank line in the logfile, not even the
// timestamp will be added to that line. Useful for separating parts of the
// logfile

{$IFDEF VER120}
// Delphi 4.x specific code
procedure Log; overload;
procedure Log(Message: string); overload;
procedure Log(const Format: string; const Args: array of const); overload;
procedure LogSeparator(Message: string; LineChar: Char='_');

// Please note that only Published properties of objects will be logged
procedure Log(Obj: TObject; MaxStringDump: Integer=10); overload;

// Binary logging
procedure Log(Name: string; var Data; Size: Integer); overload;
{$ELSE}
// Delphi 3.x specific code
procedure Log;
procedure LogText(Message: string);
procedure LogFormat(const Format: string; const Args: array of const);
procedure LogSeparator(Message: string; LineChar: Char);

// Binary logging
procedure LogBinary(Name: string; var Data; Size: Integer); 
{$ENDIF}


implementation

{$IFDEF LOGGING}

uses
	Math, Windows, Classes, Graphics, Controls, SysUtils, TypInfo;

var
	FirstLine	: Boolean	= TRUE;

procedure AssignLogFile(var LogFile: TextFile);
begin
	if (LogFilename <> '') then
		AssignFile(LogFile, LogFilename)
	else
		AssignFile(LogFile, ExtractFilePath(ParamStr(0))+'Log.txt');
end; // procedure AssignLogFile

{$IFDEF VER120}
// Delphi 4.x specific code
procedure Log(Message: string);
{$ELSE}
// Delphi 3.x specific code
procedure LogText(Message: string);
{$ENDIF}
var
	LogFile	: TextFile;
	FileRec	: TFileRec absolute LogFile;
begin
	// Open the file, possibly create it
	AssignLogFile(LogFile);
	{$I-}
	Append(LogFile);
	{$I+}
	if (IOResult <> 0) then
		ReWrite(LogFile);

	// Check for whole line
	if (Length(Message) = 3) and (Message[1] = Message[2]) and (Message[2] = Message[3]) then
		repeat
			Message := Message + Message[1];
		until (Length(Message) = LineLength);
		
	// Write message to log
	if IncludeTimeStamp and (Message <> '') then
		Write(LogFile, DateTimeToStr(Now), '    ');

	WriteLn(LogFile, Message);
	FirstLine := FALSE;

	// Flush file ?
	if AutoFlush then
		FlushFileBuffers(FileRec.Handle); 

	// Close logfile
	CloseFile(LogFile);
end; // procedure Log(string)

procedure Log;
begin
{$IFDEF VER120}
// Delphi 4.x specific code
	Log('');
{$ELSE}
// Delphi 3.x specific code
	LogText('');
{$ENDIF}
end; // procedure Log

{$IFDEF VER120}
// Delphi 4.x specific code
procedure Log(const Format: string; const Args: array of const);
{$ELSE}
// Delphi 3.x specific code
procedure LogFormat(const Format: string; const Args: array of const);
{$ENDIF}
begin
{$IFDEF VER120}
// Delphi 4.x specific code
	Log(SysUtils.Format(Format, Args));
{$ELSE}
// Delphi 3.x specific code
	LogText(SysUtils.Format(Format, Args));
{$ENDIF}
end; // procedure Log(string,array of const)

{$IFDEF VER120}
procedure Log(Obj: TObject; MaxStringDump: Integer);
type
	TDumpType	= (dtProperty, dtObject);

	procedure Log(Level: Integer; Message: string);
	begin
		// Indent message
		while (Level > 0) do
		begin
			Message := '  ' + Message;
			Dec(Level);
		end; // while (Level > 0)

		// Log it
		Logging.Log(Message);
	end; // procedure Log.Log

	procedure Dump(DumpType: TDumpType; PropertyName: string; Level: Integer; Obj: TObject);
	var
		PropList	: PPropList;
		Index			: Integer;
		PropName	: string;
		PropType	: string;
		PropValue	: string;
		PropData	: PTypeData;
		PropKind	: TTypeKind;
		SubObject	: TObject;

		procedure StartDump(DumpType: TDumpType);
		const
			DumpName	: array[TDumpType] of string = (
				'property',
				'object'
			);
		begin
			// Select correct header
			if (Obj is TControl) then
				Log(Level, DumpName[DumpType] + ' ' + (Obj as TControl).Name + ': ' +
					Obj.ClassName)
			else if (DumpType = dtProperty) then
				Log(Level, DumpName[DumpType] + ' ' + PropertyName + ': ' + Obj.ClassName)
			else
				Log(Level, DumpName[DumpType] + Obj.ClassName);
		end; // procedure Log.Dump.StartDump

		procedure EndDump;
		begin
			// End object or property
			Log(Level, 'end');
		end; // procedure Log.Dump.EndDump

		function IntegerValue: string;
		begin
			if (PropType = 'TColor') then
				Result := ColorToString(GetOrdProp(Obj, PropList[Index]))
			else
				Result := IntToStr(GetOrdProp(Obj, PropList[Index]));
		end; // function Log.Dump.IntegerValue

		function Int64Value: string;
		begin
			Result := IntToStr(GetInt64Prop(Obj, PropList[Index]));
		end; // function Log.Dump.Int64Value

		function CharValue: string;
		begin
			Result := '''' + Chr(GetOrdProp(Obj, PropList[index])) + '''';
		end; // function Log.Dump.CharValue

		function EnumerationValue: string;
		begin
			Result := GetEnumName(PropList[Index].PropType^, GetOrdProp(Obj, PropList[Index]));
		end; // function Log.Dump.EnumerationValue

		function FloatValue: string;
		begin
			Result := FloatToStr(GetFloatProp(Obj, PropList[Index]));
		end; // function Log.Dump.FloatValue

		function StringValue: string;
		begin
			Result := '''' + GetStrProp(Obj, PropList[Index]) + '''';
		end; // function Log.Dump.StringValue

		function SetValue: string;
		var
			Mask	: Integer;
			Bit		: Integer;
		begin
			// No values found yet
			Result := '';

			// Retrieve the bit mask
			Mask := GetOrdProp(Obj, PropList[Index]);

			// Loop through the bits
			for Bit := 0 to 31 do
				if ((Mask and (1 shl Bit)) <> 0) then
					Result := Result + ', ' + GetEnumName(PropData.CompType^, Bit);

			// Now fix-up the results
			if (Result <> '') then
				Delete(Result, 1, 2);
			Result := '[' + Result + ']';
		end; // function Log.Dump.SetValue

		function MethodValue: string;
		var
			Method	: TMethod;
		const
			MethodTypeName	: array[TMethodKind] of string = (
				'procedure', 'function', 'constructor', 'destructor',
				'class procedure', 'class function', 'safe procedure',
				'safe function'
			);

			function MethodName(P: Pointer): string;
			var
				C	: TClass;
			begin
				C := Obj.ClassType;

				// Loop down to TObject
				repeat
					// Found a match?
					if (C.MethodName(P) <> '') then
					begin
						Result := C.ClassName + DotSep + C.MethodName(P);
						Exit;
					end; // if (C.MethodName(P) <> '')

					// No, so go to parent
					C := C.ClassParent;
				until (C = nil); // repeat

				// No method found, return pointer string instead
				Result := '$'+IntToHex(Integer(P), 8);
			end; // function Log.Dump.MethodValue.MethodName

		begin
			// No method found yet
			Result := '';

			// Retrieve method data
			Method := GetMethodProp(Obj, PropList[Index]);

			// Check the method
			if (Method.Code = nil) then
			begin
				// Return empty name
				Exit;
			end // if (Method.Code = nil)
			else begin
				PropName := MethodTypeName[PropData.MethodKind] + ' ' + PropName;
				Result := MethodName(Method.Code);
			end;
		end; // function Log.Dump.MethodValue

		function WCharValue: string;
		begin
			Result := '#' + IntToStr(GetOrdProp(Obj, PropList[Index]));
		end; // function Log.Dump.WCharValue

		function VariantValue: string;
		var
			V			: Variant;
			S			: string;
			B			: Boolean;
			j			: Integer;
			j1,j2	: Integer;
		begin
			V := GetVariantProp(Obj, PropList[Index]);
			// Variant arrays are special
			if ((VarType(V)  and varArray) <> 0) then
			begin
				Result := '(Array) [';
				// Determine bounds
				j1 := VarArrayLowBound(V, 1);
				j2 := VarArrayHighBound(V, 1);
				j := j1;

				// Loop through at most 10 elements of the array
				while (j <= Min(j1+10, j2)) do
				begin
					S := V[j];
					if (Pos(',', S) > 0) then
						S[Pos(',', S)] := '.';
						
					if (VarType(V[j]) = varString) or (VarType(V[j]) = varOLEStr) then
						Result := Result + '''' + S + ''', '
					else
						Result := Result + S + ', ';
					Inc(j);
				end; // while (j <= Min(j1+10, j2))

				// Did we get the whole array ?
				if (j <= j2) then
					Result := Result + '..., ';
				Result := Copy(Result, 1, Length(Result)-2) + ']';
			end // if (VarType(V) = varArray)
			else begin
				S := V;
				case (VarType(V) and (not varByRef)) of
					varEmpty		: Result := '(Empty)';
					varNull			: Result := '(Null)';
					varSmallInt	: Result := '(SmallInt) ' + S;
					varInteger	: Result := '(Integer) ' + S;
					varSingle		: Result := '(Single) ' + S;
					varDouble		: Result := '(Double)' + S;
					varCurrency	: Result := '(Currency) ' + S;
					varDate			: Result := '(Date) ' + S;
					varOLEStr		: Result := '(OLEStr) ' + S;
					varDispatch	: Result := '(Dispatch)';
					varError		: Result := '(Error) ' + S;
					varBoolean	: begin
						B := V;
						Result := '(Boolean) ' + BooleanIdents[B];
					end; // case VarType(V) of varBoolean
					varUnknown	: Result := '(Unknown)';
					varByte			: Result := '(Byte) ' + S;
					varString		: Result := '(String) ''' + S + '''';
					varTypeMask	: Result := '(TypeMask) ' + S;
					varArray		: Result := '(Array) ' + S;
					varByRef		: Result := '(ByRef) ' + S;
				end; // case (VarType(V) and (not varByRef))
				if (Pos(',', S) > 0) then
					S[Pos(',', S)] := '.';
			end; // if (VarType(V) = varArray) ... else
		end; // function Log.Dump.WCharValue

		procedure DumpStrings;
		var
			Index	: Integer;
			S			: string;
		begin
			if ((Obj as TStrings).Count > 0) then
			begin
				Log(Level + 1, 'items = (');
				for Index := 1 to Max(1, Min((Obj as TStrings).Count, MaxStringDump)) do
				begin
					S := '''' + (Obj as TStrings)[Index-1] + '''';
					if (Index < (Obj as TStrings).Count) then
						S := S + ',';
					Log(Level + 2, S);
				end; // for Index := 1 to (Obj as TStrings).Count
				Log(Level + 1, ')');
			end; // if ((Obj as TStrings).Count > 0)
		end; // procedure Log.Dump.DumpStrings

		function HasProperties(Obj: TObject): Boolean;
		var
			PropList	: PPropList;
		begin
			// Allocate memory for property list
			New(PropList);

			// Initialize and retrieve property list
			FillChar(PropList^, SizeOf(PropList^), 0);
			GetPropInfos(Obj.ClassInfo, PropList);

			// If it got properties, return it
			Result := Assigned(PropList[0]);

			// Deallocate memory from property list
			Dispose(PropList);
		end; // function Log.Dump.HasProperties 

	begin
		// Log header of dump
		StartDump(DumpType);

		// Allocate memory for property list
		New(PropList);

		// Is this a TStrings descendant ?
		if (Obj is TStrings) then
			DumpStrings;

		// Initialize and retrieve property list
		FillChar(PropList^, SizeOf(PropList^), 0);
		GetPropInfos(Obj.ClassInfo, PropList);

		// Now process all properties for this class
		for Index := Low(PropList^) to High(PropList^) do
		begin
			// If we have reached the end, break th eloop
			if (not Assigned(PropList[Index])) then
				Break;

			PropName := PropList[Index].Name;
			PropType := PropList[Index].PropType^.Name;
			PropData := GetTypeData(PropList[Index].PropType^);
			PropKind := PropList[Index].PropType^.Kind;

			if (PropKind = tkClass) then
			begin
				SubObject := TObject(GetOrdProp(Obj, PropList[Index]));
				if Assigned(SubObject) then
				begin
					if (SubObject is TIcon) then
					begin
						if TIcon(SubObject).Empty then
							Log(Level + 1, PropName + ' = (none)')
						else
							Log(Level + 1, PropName + ' = (TIcon)');
					end // if (SubObject.ClassName = 'TIcon')
					else if (SubObject is TStrings) then
					begin
						if ((SubObject as TStrings).Count > 0) then
							Dump(dtProperty, PropName, Level + 1, SubObject)
						else
							Log(Level + 1, PropName + ' = (' + SubObject.ClassName + ')');
					end // if (SubObject is TStrings)
					else if HasProperties(SubObject) then
						Dump(dtProperty, PropName, Level + 1, SubObject)
					else
						Log(Level + 1, PropName + ' = (' + SubObject.ClassName + ')');
				end; // if Assigned(SubObject)
			end // if (PropKind = tkClass)
			else begin
				case PropKind of
					tkUnknown			: PropValue := '?';
					tkInteger			: PropValue := IntegerValue;
					tkInt64				: PropValue := Int64Value;
					tkChar				: PropValue := CharValue;
					tkEnumeration	: PropValue := EnumerationValue;
					tkFloat				: PropValue := FloatValue;
					tkString			: PropValue := StringValue;
					tkWString			: PropValue := StringValue;
					tkLString			: PropValue := StringValue;
					tkSet					: PropValue := SetValue;
					tkMethod			: begin
						PropValue := MethodValue;
						if (PropValue = '') then
							Continue;
					end; // class PropKind of tkMethod
					tkWChar				: PropValue := WCharValue;
					tkVariant			: PropValue := VariantValue;
					tkArray				: PropValue := 'array[] of ?'; 		// Cannot be published
					tkRecord			: PropValue := 'record';					// Not found as published
					tkInterface		: PropValue := 'interface';
					tkDynArray		: PropValue := 'array of ?';			// Cannot be published
				end; // case PropList[Index].PropType^.Kind

				// Now log the data
				Log(Level + 1, PropName + ' = ' + PropValue);
			end // if (PropKind = tkClass) ... else
		end; // for Index := Low(PropList^) to High(PropList^)

		// Deallocate memory from property list
		Dispose(PropList);

		// Check if this owns other objects
		if (Obj is TWinControl) then
		begin
			for Index := 1 to (Obj as TWinControl).ControlCount do
				Dump(dtObject, '', Level + 1, (Obj as TWinControl).Controls[Index-1]);
		end; // if (Obj is TWinControl)

		// Log footer of dump
		EndDump;
	end; // procedure Log.Dump

begin
	Dump(dtObject, '', 0, Obj);
end; // procedure Log(TObject, Integer)
{$ENDIF}

procedure EmptyLog;
var
	LogFile	: TextFile;
begin
	AssignLogFile(LogFile);
	{$I-}
	Erase(LogFile);
	{$I+}
	if (IOResult <> 0) then
		; // Just to clear IOResult
end; // procedure EmptyLog

procedure LogSeparator(Message: string; LineChar: Char);
begin
	// Separate by a blank line
	if (not FirstLine) then
		Log;

	// Pad to 80 chars
	Message := LineChar + LineChar + LineChar + ' ' + Message + ' ';
	while (Length(Message) < LineLength) do
		Message := Message + LineChar;

	// Log message
{$IFDEF VER120}
// Delphi 4.x specific code
	Log(Message);
{$ELSE}
// Delphi 3.x specific code
	LogText(Message);
{$ENDIF}
end; // procedure LogSeparator

{$IFDEF VER120}
// Delphi 4.x specific code
procedure Log(Name: string; var Data; Size: Integer);
{$ELSE}
// Delphi 3.x specific code
procedure LogBinary(Name: string; var Data; Size: Integer);
{$ENDIF}
var
	i,j		: Integer;
	Pos		: Integer;
	Line	: string;
	P			: PChar;
begin
	if (Size > 0) then
	begin
		P := @Data;
{$IFDEF VER120}
// Delphi 4.x specific code
		Log('Binary dump: '+Name);
{$ELSE}
// Delphi 3.x specific code
		LogText('Binary dump: '+Name);
{$ENDIF}
		Line := '                                                 -                  ';
		for i := 1 to Size do
		begin
			j := (i-1) mod 16;
			Pos := j * 3 + 1;
			Line := Copy(Line, 1, Pos-1) + IntToHex(Ord(P[j]), 2) + Copy(Line, Pos+2, Length(Line));

			Pos := 53 + j;
			if (P[j] in [#32..#255]) then
				Line[Pos] := P[j]
			else
				Line[Pos] := '.';
			if (i mod 16) = 0 then
{$IFDEF VER120}
// Delphi 4.x specific code
				Log(Line);
{$ELSE}
// Delphi 3.x specific code
				LogText(Line);
{$ENDIF}
		end; // for i := 1 to Size
		if (Size mod 16) <> 0 then
{$IFDEF VER120}
// Delphi 4.x specific code
			Log(Line);
{$ELSE}
// Delphi 3.x specific code
			LogText(Line);
{$ENDIF}
	end; // if (i > 0)
end; // procedure Log(string, var, Integer)

{$IFDEF BLANKLOG}
initialization
	EmptyLog;
{$ENDIF}		// {$IFDEF BLANKLOG}
{$ELSE}			// {$IFDEF LOGGING}

procedure Log(Message: string);
begin
end; // procedure Log(string)

procedure Log;
begin
end; // procedure Log

procedure Log(const Format: string; const Args: array of const);
begin
end; // procedure Log(string, array of const)

procedure Log(Obj: TObject; MaxStringDump: Integer=10);
begin
end; // procedure Log(TObject, Integer)

procedure LogSeparator(Message: string; LineChar: Char='*');
begin
end; // procedure LogSeparator

procedure Log(Name: string; var Data; Size: Integer);
begin
end; // procedure Log(string, var, Integer)

{$ENDIF}		// {$IFDEF LOGGING ... ELSE}
end.

