//
// $History: RC5.pas $
// 
// *****************  Version 6  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
// 
// *****************  Version 3  *****************
// User: Lvk          Date: 22.07.98   Time: 17:21
// Updated in $/Delphi 4.0/StreamFilter
// Fixed some comment blocks
// 
// *****************  Version 2  *****************
// User: Lvk          Date: 21.07.98   Time: 23:52
// Updated in $/Delphi 4.0/StreamFilter
// Made the base class (TCryptography) expand the key and make a binary
// key of the correct size, thereby reducing code in each descendant
// class.
// 
// Also fixed some minor bugs and discrepancies in the comments
// 
// *****************  Version 1  *****************
// User: Lvk          Date: 20.07.98   Time: 20:51
// Created in $/Delphi 4.0/StreamFilter
// RC5 encryption algorithm
//

unit RC5;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net
//
// The RC5 algorithm was developed by Ronald L. Rivest,
// MIT Laboratory for Computer Science
// rivest@theory.lcs.mit.edu
//
// Paper describing this algorithm was downloaded from
// ftp://ftp.rsa.com/pub/rc5/rc5.ps

interface

uses
	AuxFuncs, SysUtils, Cryptography;

const
	Rounds		= 12;
	KeyLength	= 2 * (Rounds + 1);

type
	TRC5Block	= array[1..2] of LongWord;

{ TRC5 class }
	TRC5 = class(TCryptography)
	private
		S	: array[0..KeyLength-1] of LongWord;
	protected
		procedure CalculateSubKeys; override;
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create; overload;
		constructor Create(AKey: string); overload;
{$ELSE}
// Delphi 3.x specific code
		constructor Create;
		constructor CreateKey(AKey: string);
{$ENDIF}
		destructor Destroy; override;

		function EncipherBlock(var Block): Boolean; override;
		function DecipherBlock(var Block): Boolean; override;
	end; // class TRC5

implementation

{ TRC5 }

const
	P32	= $b7e15163;
	Q32	= $9e3779b9;

{$R-,Q-}
procedure TRC5.CalculateSubKeys;
var
	i,j,k	: Integer;
	L			: array[0..15] of LongWord;
	A,B		: LongWord;
begin
	// Copy the key into L
	Move(KeyPtr^, L, KeySize);

	// Now initialize the S table
	S[0] := P32;
	for i := 1 to KeyLength-1 do
		S[i] := S[i-1] + Q32;

	// Now scramble the S table with the key
	i := 0;
	j := 0;
	A := 0;
	B := 0;
	for k := 1 to 3*KeyLength do
	begin
		A := ROL((S[i] + A + B), 3);
		S[i] := A;
		B := ROL((L[j] + A + B), (A + B));
		L[j] := B;
		i := (i + 1) mod KeyLength;
		j := (j + 1) mod 16;
	end; // for k := 1 to 3*KeyLength
end; // procedure TRC5.CalculateSubKeys
{$R+,Q+}

{$IFDEF VER120}
// Delphi 4.x specific code
constructor TRC5.Create(AKey: string);
begin
	inherited Create('RC5', 8, 16 * 4, AKey);
end; // constructor TRC5.Create(string)
{$ELSE}
// Delphi 3.x specific code
constructor TRC5.CreateKey(AKey: string);
begin
	inherited CreateKey('RC5', 8, 16 * 4, AKey);
end; // constructor TRC5.Create(string)
{$ENDIF}

constructor TRC5.Create;
begin
	inherited Create('RC5', 8, 16 * 4);
end; // constructor TRC5.Create

{$R-,Q-}
function TRC5.DecipherBlock(var Block): Boolean;
var
	RC5Block	: TRC5Block absolute Block;
	i					: Integer;
begin
	for i := Rounds downto 1 do
	begin
		RC5Block[2] := ROR((RC5Block[2]-S[2*i+1]), RC5Block[1]) xor RC5Block[1];
		RC5Block[1] := ROR((RC5Block[1]-S[2*i]), RC5Block[2]) xor RC5Block[2];
	end; // for i := 1 to Rounds

	Dec(RC5Block[2], S[1]);
	Dec(RC5Block[1], S[0]);

	Result := TRUE;
end; // function TRC5.DecipherBlock
{$R+,Q+}

destructor TRC5.Destroy;
begin
	inherited;
end; // destructor TRC5.Destroy

{$R-,Q-}
function TRC5.EncipherBlock(var Block): Boolean;
var
	RC5Block	: TRC5Block absolute Block;
	i					: Integer;
begin
	Inc(RC5Block[1], S[0]);
	Inc(RC5Block[2], S[1]);

	for i := 1 to Rounds do
	begin
		RC5Block[1] := ROL((RC5Block[1] xor RC5Block[2]),RC5Block[2]) + S[2*i];
		RC5Block[2] := ROL((RC5Block[2] xor RC5Block[1]),RC5Block[1]) + S[2*i+1];
	end; // for i := 1 to Rounds

	Result := TRUE;
end; // function TRC5.EncipherBlock
{$R+,Q+}

end. // unit RC5
