//
// $History: RC6.pas $
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 3  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
// 
// *****************  Version 2  *****************
// User: Lvk          Date: 22.07.98   Time: 17:21
// Updated in $/Delphi 4.0/StreamFilter
// Fixed some comment blocks
// 
// *****************  Version 1  *****************
// User: Lvk          Date: 21.07.98   Time: 23:54
// Created in $/Delphi 4.0/StreamFilter
// RC6 Encryption Algorithm
//

unit RC6;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net
//
// The RC6 algorithm was developed by Ronald L. Rivest,
// M.J.B. Robshaw, R. Sidney, and Y.L. Yin
//
// Paper describing this algorithm was downloaded from
// http://theory.lcs.mit.edu/~rivest/rc6.pdf

interface

uses
	AuxFuncs, Cryptography;

const
	Rounds		= 20;
	KeyLength	= 2 * (Rounds + 2);

type
	TRC6Block	= array[1..4] of LongWord;

{ TRC6 class }
	TRC6 = class(TCryptography)
	private
		S	: array[0..KeyLength-1] of LongWord;
	protected
		procedure CalculateSubKeys; override;
	public
{$IFDEF VER120}
// Delphi 4.x specific code
		constructor Create; overload;
		constructor Create(AKey: string); overload;
{$ELSE}
// Delphi 3.x specific code
		constructor Create;
		constructor CreateKey(AKey: string);
{$ENDIF}
		destructor Destroy; override;

		function EncipherBlock(var Block): Boolean; override;
		function DecipherBlock(var Block): Boolean; override;
	end; // class TRC6

implementation

{ TRC6 }

const
	P32	= $b7e15163;
	Q32	= $9e3779b9;
	lgw	= 5;					// log(32,2)

// Key expansion used as for RC5, since no explanation or description of
// any changes was shown in the RC6 document.
procedure TRC6.CalculateSubKeys;
var
	i,j,k	: Integer;
	L			: array[0..15] of LongWord;
	A,B		: LongWord;
begin
	// Copy the key into L
	Move(KeyPtr^, L, KeySize);

	// Now initialize the S table
	S[0] := P32;
	for i := 1 to KeyLength-1 do
		S[i] := S[i-1] + Q32;

	// Now scramble the S table with the key
	i := 0;
	j := 0;
	A := 0;
	B := 0;
	for k := 1 to 3*KeyLength do
	begin
		A := ROL((S[i] + A + B), 3);
		S[i] := A;
		B := ROL((L[j] + A + B), (A + B));
		L[j] := B;
		i := (i + 1) mod KeyLength;
		j := (j + 1) mod 16;
	end; // for k := 1 to 3*KeyLength
end; // procedure TRC6.CalculateSubKeys

{$IFDEF VER120}
// Delphi 4.x specific code
constructor TRC6.Create(AKey: string);
begin
	inherited Create('RC6', SizeOf(TRC6Block), 16 * 4, AKey);
end; // constructor TRC6.Create(string)
{$ELSE}
// Delphi 3.x specific code
constructor TRC6.CreateKey(AKey: string);
begin
	inherited CreateKey('RC6', SizeOf(TRC6Block), 16 * 4, AKey);
end; // constructor TRC6.Create(string)
{$ENDIF}

constructor TRC6.Create;
begin
	inherited Create('RC6', SizeOf(TRC6Block), 16 * 4);
end; // constructor TRC6.Create

function TRC6.DecipherBlock(var Block): Boolean;
var
	RC6Block	: TRC6Block absolute Block;
	i					: Integer;
	t,u				: LongWord;
	Temp			: LongWord;
begin
	// Init the block
	RC6Block[3] := RC6Block[3] - S[2*Rounds+3];
	RC6Block[1] := RC6Block[1] - S[2*Rounds+2];

	// Loop through all S elements
	for i := Rounds downto 1 do
	begin
		// Rotate the words
		Temp := RC6Block[4];
		RC6Block[4] := RC6Block[3];
		RC6Block[3] := RC6Block[2];
		RC6Block[2] := RC6Block[1];
		RC6Block[1] := Temp;

		// Do the math
		u := ROL((RC6Block[4] * (2*RC6Block[4] + 1)),lgw);
		t := ROL((RC6Block[2] * (2*RC6Block[2] + 1)),lgw);
		RC6Block[3] := ROR((RC6Block[3]-S[2*i+1]), t) xor u;
		RC6Block[1] := ROR((RC6Block[1]-S[2*i]), u) xor t;
	end; // for i := 1 to Rounds

	// Adjust for final round
	Dec(RC6Block[4], S[1]);
	Dec(RC6Block[2], S[0]);

	// This function cannot fail
	Result := TRUE;
end; // function TRC6.DecipherBlock

destructor TRC6.Destroy;
begin
	inherited;
end; // destructor TRC6.Destroy

function TRC6.EncipherBlock(var Block): Boolean;
var
	RC6Block	: TRC6Block absolute Block;
	i					: Integer;
	t,u				: LongWord;
	Temp			: LongWord;
begin
	// Init block
	Inc(RC6Block[2], S[0]);
	Inc(RC6Block[4], S[1]);

	// Loop through all S elements
	for i := 1 to Rounds do
	begin
		// Do the math
		t := ROL((RC6Block[2] * (2*RC6Block[2] + 1)),lgw);
		u := ROL((RC6Block[4] * (2*RC6Block[4] + 1)),lgw);
		RC6Block[1] := ROL((RC6Block[1] xor t), u) + S[2*i];
		RC6Block[3] := ROL((RC6Block[3] xor u), t) + S[2*i+1];

		// Rotate the words
		Temp := RC6Block[1];
		RC6Block[1] := RC6Block[2];
		RC6Block[2] := RC6Block[3];
		RC6Block[3] := RC6Block[4];
		RC6Block[4] := Temp;
	end; // for i := 1 to Rounds

	// Adjust for final round
	RC6Block[1] := RC6Block[1] + S[2*Rounds+2];
	RC6Block[3] := RC6Block[3] + S[2*Rounds+3];

	// This function cannot fail
	Result := TRUE;
end; // function TRC6.EncipherBlock

end. // unit RC6
