//
// $History: SHA.pas $
// 
// *****************  Version 7  *****************
// User: Lvk          Date: 20.10.98   Time: 23:37
// Updated in $/Delphi 4.0/StreamFilter
// Modified email address + added persistent storage classes
// 
// *****************  Version 6  *****************
// User: Lvk          Date: 12.09.98   Time: 17:36
// Updated in $/Delphi 4.0/StreamFilter
// 
// *****************  Version 5  *****************
// User: Lvk          Date: 9.09.98    Time: 19:15
// Updated in $/Delphi 4.0/StreamFilter
// - Made library Delphi 3.x compatible
// - Fixed bug in encryption base classes
// - Replaced old email address
// 
// *****************  Version 4  *****************
// User: Lvk          Date: 22.07.98   Time: 17:21
// Updated in $/Delphi 4.0/StreamFilter
// Fixed some comment blocks
// 
// *****************  Version 3  *****************
// User: Lvk          Date: 20.07.98   Time: 20:50
// Updated in $/Delphi 4.0/StreamFilter
// Added documentation for first release
// 
// *****************  Version 2  *****************
// User: Lvk          Date: 20.07.98   Time: 18:11
// Updated in $/Delphi 4.0/StreamFilter
// Fixed bugs in the digest calculation code
//
// *****************  Version 1  *****************
// User: Lvk          Date: 19.07.98   Time: 23:50
// Created in $/Delphi 4.0/StreamFilter
// Secure Hash Algorithm (SHA) CheckSum class
//

unit SHA;

// (C) Lasse Vgsther Karlsen, 1998
// lasse.karlsen@c2i.net
//
// Secure Hash Algorithm (SHA) was developed by the National Institute of
// Standards and Technology (NIST) and was subsequently published as a
// federal information-processing standard (FIPS PUB 180) in 1993.
//
// It is based on the MD4 algorithm, developed by Ron Rivest.

interface

uses
	AuxFuncs, Math, CheckSum;

type
	TSignature	= array[0..4] of LongWord;
	TBlock			= array[1..64] of Byte;

{ TSHA class }
	TSHA	= class(TBasicCheckSum)
	private
		FSignature						: TSignature;
		FSignatureCalculated	: Boolean;

		FTempSignature				: TSignature;
		FBlock								: TBlock;
		FInBlock							: Integer;
{$IFDEF VER120}
// Delphi 4.x specific code
		FLength								: Int64;
{$ELSE}
// Delphi 3.x specific code
		FLength								: Integer;
{$ENDIF}
	protected
		procedure CalculateSignature; virtual;
		procedure FlushBlock; virtual;
		function GetByte(Index: Integer): Byte; override;
	public
		constructor Create;
		procedure Reset; override;
		procedure Process(const Data; Size: Integer); override;
	end; // class TSHA

implementation

resourcestring
	SSHA	= 'Secure Hash Algorithm (SHA)';

{ TSHA }

procedure TSHA.CalculateSignature;
var
	DummyBlock		: array[0..63] of Byte;
	DummyByte			: Byte;
	OldBlock			: TBlock;
	OldInBlock		: Integer;
	OldSignature	: TSignature;
{$IFDEF VER120}
// Delphi 4.x specific code
	OldLength			: Int64;
	Temp					: Int64;
{$ELSE}
// Delphi 3.x specific code
	OldLength			: Integer;
	Temp					: packed record
		Length			: Integer;
		Zeroes			: Integer;
	end;
{$ENDIF}
begin
	// No need to recalculate if its already done
	if FSignatureCalculated then
		Exit;

	// Save old data before doing this
	OldBlock := FBlock;
	OldInBlock := FInBlock;
	OldLength := FLength;
	OldSignature := FTempSignature;

	// We need to pad with 1 1-bit and the rest 0-bits
	DummyByte := 128;
	FillChar(DummyBlock, SizeOf(DummyBlock), 0);

	// Now pad the data
	Process(DummyByte, 1);
	repeat
		if (FInBlock >= 56) then
			Process(DummyBlock, 64-FInBlock)
		else
			Process(DummyBlock, 56-FInBlock);
	until (FInBlock = 56); // repeat

	// And append the length
{$IFDEF VER120}
// Delphi 4.x specific code
	Temp := SwapHiLo64(OldLength);
	Process(Temp, SizeOf(OldLength));
{$ELSE}
// Delphi 3.x specific code
	Temp.Length := OldLength;
	Temp.Zeroes := 0;
	SwapHiLo64(Temp);
	Process(Temp, SizeOf(Temp));
{$ENDIF}

	// Copy the signature
	FSignature := FTempSignature;

	// Restore old data
	FBlock := OldBlock;
	FInBlock := OldInBlock;
	FLength := OldLength;
	FTempSignature := OldSignature;
	FSignatureCalculated := TRUE;
end; // procedure TSHA.CalculateSignature

constructor TSHA.Create;
begin
	inherited Create(SSHA, 20);
end; // constructor TSHA.Create

{$R-,Q-}
procedure TSHA.FlushBlock;
var
	Round			: Integer;
	A,B,C,D,E	: LongWord;
	Temp			: LongWord;
	W					: array[0..79] of LongWord;
	i					: Integer;
const
	K	: array[0..3] of LongWord = (
		$5a827999, $6ed9eba1, $8f1bbcdc, $ca62c1d6
	);

	function F(B,C,D: LongWord): LongWord;
	begin
		Result := 0;
		case Round of
			0..19		: Result := (B and C) or ((not B) and D);
			20..39	: Result := B xor C xor D;
			40..59	: Result := (B and C) or (B and D) or (C and D);
			60..79	: Result := B xor C xor D;
		end; // case Round
	end; // function TSHA.FlushBlock.F

begin
	// Calculate W[]
	Move(FBlock, W, SizeOf(FBlock));
	for i := 0 to 15 do
		W[i] := SwapHiLo32(W[i]);
		
	for i := 16 to 79 do
	begin
		W[i] := ROL(W[i-16] xor W[i-14] xor W[i-8] xor W[i-3], 1);
	end; // for i := 16 to 79

	// Put old signature values in A-E
	A := FTempSignature[0];
	B := FTempSignature[1];
	C := FTempSignature[2];
	D := FTempSignature[3];
	E := FTempSignature[4];

	// Do the math
	for Round := 0 to 79 do
	begin
		Temp := ROL(A, 5) + E + K[Round div 20] + W[Round] + F(B,C,D);
		E := D;
		D := C;
		C := ROL(B, 30);
		B := A;
		A := Temp;
	end; // for Round := 0 to 79

	// Now add it to the signature
	Inc(FTempSignature[0], A);
	Inc(FTempSignature[1], B);
	Inc(FTempSignature[2], C);
	Inc(FTempSignature[3], D);
	Inc(FTempSignature[4], E);

	// Reset data
	FillChar(FBlock, SizeOf(FBlock), 0);
	FInBlock := 0;
end; // procedure TSHA.FlushBlock
{$R+,Q+}

function TSHA.GetByte(Index: Integer): Byte;
var
	I	: Integer;
begin
	// Check if signature is calculated
	if not FSignatureCalculated then
		CalculateSignature;

	// Find the correct byte
	I := (Index div 4);
	case (Index mod 4) of
		0	: Result := FSignature[I] shr 24;
		1	: Result := (FSignature[I] shr 16) and 255;
		2	: Result := (FSignature[I] shr 8) and 255;
		3	: Result := (FSignature[I] and 255);
	else
		Result := 0;
	end; // case (Index mod 4)
end; // function TSHA.GetByte

procedure TSHA.Process(const Data; Size: Integer);
var
	P				: PChar;
	ToCopy	: Integer;
begin
	P := @Data;
	while (Size > 0) do
	begin
		// Move data to the temporary block
		ToCopy := Min(64 - FInBlock, Size);
		Move(P^, FBlock[FInBlock + 1], ToCopy);
		Inc(FInBlock, ToCopy);
		Inc(P, ToCopy);
		Dec(Size, ToCopy);
		Inc(FLength, ToCopy*8);

		// See if we got one whole block now
		if (FInBlock = 64) then
			FlushBlock;
	end; // while (Size > 0)
end; // procedure TSHA.Process

procedure TSHA.Reset;
const
	InitSignature	: TSignature = (
		$67452301, $efcdab89, $98badcfe, $10325476, $c3d2e1f0
	);
begin
	FillChar(FBlock, SizeOf(FBlock), 0);
	FInBlock := 0;
	FSignatureCalculated := FALSE;
	FLength := 0;

	FSignature := InitSignature;
	FTempSignature := InitSignature;
end; // procedure TSHA.Reset

end. // unit SHA
